#!/usr/bin/env python

# Usage: python -i tune.py <role> <element>
#
# See http://pulseaudio.org/wiki/DBusInterface for pulseaudio's D-Bus API
# documentation.

import sys
import dbus
from dbus.mainloop.glib import DBusGMainLoop
import gobject
import os
import subprocess
import re
import time

import readline

from pulseaudio.conversions import sw2db
from pulseaudio.conversions import db2sw


# This is just a convenience struct
# to store global data for this interpreter
# session.
class Userdata:
    def __init__(self):
        self.conn = None                    # connection to pa dbus interface
        self.core = None
        self.stream_restore = None
        self.stream_restore_entry = None
        self.stream_restore_role = None     # fulle stream-restore role name as string
        self.entry_path = None              # stream-restore entry path as string
        self.media_modes = None             # dict containing media modes
        self.phone_modes = None             # dict containing phone modes
        self.modes = None                   # dict containing previous dicts
        self.playback_r = None              # regex for parsing alsa mixer element values
        self.current_role = None            # current role as string
        self.current_mode = None            # current mode as string
        self.sweep_step = 100               # default volume step in operations
        self.sweep_sleep = None             # float value for sleeping, None no sleep
        self.watch_volume = 0

class AlsaMixerElement:
    def __init__(self, card_id, element_name):
        self.card_id = card_id
        self.element_name = element_name

userdata = Userdata()

# regex for parsing alsa mixer element values, group(1) is value, group(2) is dB-value
userdata.playback_r = re.compile(r'^.*Playback +([0-9]+).*\[(.*)dB\]')

AMIXER = "/usr/bin/amixer"

SERVER_ADDRESS = "unix:path=/var/run/pulse/dbus-socket"
ENTRY_NAME_BASE = "sink-input-by-media-role:"

# 100% volume in Pulseaudio's native volume units.
VOLUME_NORM = 0x10000
PA_NATIVE_0DB = int(VOLUME_NORM)

userdata.media_modes = {'ihf'       : AlsaMixerElement(0, "DAC2 Analog"),
                        'lineout'   : AlsaMixerElement(1, "TPA6140A2 Headphone")
}

userdata.phone_modes = {'ihf'       : AlsaMixerElement(0, "DAC2 Analog"),
                        'lineout'   : AlsaMixerElement(1, "TPA6140A2 Headphone"),
                        'hp'        : AlsaMixerElement(0, "DAC2 Analog")
}

userdata.modes = {'x-maemo': userdata.media_modes,
                  'phone': userdata.phone_modes}


# Since the server address is hardcoded, connecting is very simple, but when
# the run-time address discovery logic gets fixed, this should be changed too.
def connect_dbus():
    return dbus.connection.Connection(SERVER_ADDRESS)

def usage():
    string="""
Usage: You can start this script with command line parameters, for example
python -i %s x-maemo ihf

That way x-maemo role and ihf element are selected and connection to role is made upon startup.

Commands:
    set_role(ROLE)              ROLE can be x-maemo or phone, connect to stream-restore entry
    set_element(ELEMENT)        Select element for role, ihf, lineout, hp, etc.
    stp(VALUE)                  Increment/decrement current volume by VALUE, if VALUE is integer,
                                native pulseaudio volume is changed, if value is float, it is
                                translated from dB to pulseaudio volume first.
    vol(VALUE)                  Same as stp, but absolute VALUE is set.
    mute()                      Set volume to zero.
    max()                       Set volume to max (%d).
    p()                         Print current volume.
    sweep(START, STOP, STEP, FULL) Go through volume (defaults: START=0 STOP=%d, STEP=%d, FULL=False).
                                When FULL is True all volume changes are printed, otherwise only when
                                alsa mixer element step changes.
    set_sleep(VALUE)            Set sleep time period (float) between sweep steps, None to disable.
    watch()                     Watch volume changes and print when change detected.

PLEASE NOTE! Some stream must be playing on device for ALSA mixer element changes
to happen (at least pacat /dev/zero).

""" % (sys.argv[0],
       PA_NATIVE_0DB,
       PA_NATIVE_0DB,
       userdata.sweep_step)
    print string

def run_process(parameters):
    p = subprocess.Popen( parameters,
                          stdin = subprocess.PIPE,
                          stdout = subprocess.PIPE,
                          close_fds = True)
    stdout_text, stderr_text = p.communicate()
    return stdout_text.rstrip()

def set_element(element):
    m = userdata.modes.get(userdata.current_role, None)
    e = None
    if m:
        e = m.get(element, None)
    if e:
        userdata.current_element = element
        print "element = %s" % element
    else:
        print "element with name %s was not found." % element

def get_amixer_value():
    return get_amixer_value_full(userdata.current_role, userdata.current_element)

# execute amixer and get current mixer element values.
# returns tuple (int, float), (element_value, element_dB_value)
def get_amixer_value_full(mode, name):
    elem = userdata.modes.get(mode).get(name)
    mixer_output = run_process((AMIXER,
                    "-c%d" % elem.card_id,
                    "sget",
                    elem.element_name))

    for i in mixer_output.splitlines():
        m = userdata.playback_r.match(i)
        if m:
            return (int(m.group(1)), float(m.group(2)))

    print "WARNING: regex did not match amixer output"
    return (None, None)

def get_volume_dbus(entry):
    return entry.Get("org.PulseAudio.Ext.StreamRestore1.RestoreEntry", "Volume", dbus_interface="org.freedesktop.DBus.Properties")

def set_volume_dbus(entry, vol_native):
    current_vol = get_volume_dbus(entry)

    # We will copy current_vol to new_vol, with the channel map unchanged, and
    # volume set to vol_native.
    new_vol = dbus.Array()

    for (chan, vol) in current_vol:
        # The dbus.Struct constructor needs an iterable collection of data. We use
        # a tuple to hold the two integers.
        struct_data = (dbus.UInt32(chan), dbus.UInt32(vol_native))
        new_vol.append(dbus.Struct(struct_data))

    # The new value given to the Set method needs to be wrapped in a variant, but
    # that doesn't show in this code, because the Python bindings do the wrapping
    # automatically.
    entry.Set("org.PulseAudio.Ext.StreamRestore1.RestoreEntry", "Volume", new_vol, dbus_interface="org.freedesktop.DBus.Properties")

def note():
    string = """Hello. This script is just a pile of functions intended to be run inside python interpreter.
So if you just reached command line without anything useful (not counting this message), please start
this script again in interpreter mode: \n
 $ python -i %s""" % sys.argv[0]
    print string

def main():

    note()
    if len(sys.argv) > 1 and ( sys.argv[1] == "-h" or sys.argv[1] == "--help" ):
        sys.argv.pop(1)
        usage()
    else:
        connect()

def connect():
    # let's connect to pulseaudio already

    # We integrate with the GLib main loop implementation. That's the easiest way
    # to receive D-Bus signals asynchronously in Python.
    DBusGMainLoop(set_as_default=True)

    # Get the dbus.connection.Connection object.
    userdata.conn = connect_dbus()

    # Use the Python D-Bus bindings magic to create a proxy object for the central
    # core object of Pulseaudio.
    userdata.core = userdata.conn.get_object(object_path="/org/pulseaudio/core1")

    # Create a proxy object for the stream restore extension's central object.
    userdata.stream_restore = userdata.conn.get_object(object_path="/org/pulseaudio/stream_restore1")

    if len(sys.argv) > 1:
        set_role(sys.argv[1])

    if len(sys.argv) > 2:
        set_element(sys.argv[2])

def set_role(role):
    userdata.stream_restore_role = ENTRY_NAME_BASE + role
    print "role = %s" % userdata.stream_restore_role
    # Use the proxy object to transparently call GetEntryByName on the remote
    # object. Convenience has its drawbacks - this call blocks until the reply (or
    # timeout) arrives. Note that D-Bus objects can implement multiple interfaces,
    # and that's why we always specify the dbus_interface keyword argument in the
    # proxy method calls. (You could also create dbus.Interface objects, but that
    # doesn't really simplify things if you make only one call for each interface,
    # like we do here.)
    userdata.entry_path = userdata.stream_restore.GetEntryByName(userdata.stream_restore_role,
                                                                 dbus_interface="org.PulseAudio.Ext.StreamRestore1")
    # Create a proxy object for the role restore entry object.
    userdata.stream_restore_entry = userdata.conn.get_object(object_path=userdata.entry_path)

    userdata.current_role = role

    if userdata.stream_restore_entry:
        print "connected to stream-restore entry"

def set_step(step):
    userdata.sweep_step = step

def mute():
    vol(0)

def max():
    vol(PA_NATIVE_0DB)

def print_volume(volume, alsa_value, alsa_db):
    if volume > 0:
        db = sw2db(volume)
    else:
        db = -99999.0
    print "Native: %d (%f dB)   ALSA mixer: %d  (%f dB)" % (volume, db, alsa_value, alsa_db)


def s(new_vol):
    set_volume_dbus(userdata.stream_restore_entry, new_vol)
    alsa_value, alsa_db = get_amixer_value()
    print_volume(new_vol, alsa_value, alsa_db)

def get_volume():
    return volume_max(get_volume_dbus(userdata.stream_restore_entry))

def volume_max(volume):
    max_vol = 0
    for channel, vol in volume:
        if vol > max_vol:
            max_vol = vol
    return max_vol

def p():
    volume = get_volume()
    alsa_value, alsa_db = get_amixer_value()
    print_volume(volume, alsa_value, alsa_db)

def watch():
    userdata.watch_volume = -1
    userdata.conn.add_signal_receiver(watch_signal_cb, message_keyword="msg")
    userdata.core.ListenForSignal("org.PulseAudio.Ext.StreamRestore1.RestoreEntry.VolumeUpdated",
                                  [userdata.entry_path],
                                  dbus_interface="org.PulseAudio.Core1")
    loop = gobject.MainLoop()
    # ugly as hell, but oh well
    try:
        loop.run()
    except: # catch all, including KeyboardInterrupt
        print "\nEnd watch()"
        loop.quit()
        userdata.conn.remove_signal_receiver(watch_signal_cb)

def watch_signal_cb(*args, **keywords):
    s = keywords["msg"]

    if s.get_path() == userdata.entry_path \
        and s.get_interface() ==  "org.PulseAudio.Ext.StreamRestore1.RestoreEntry" \
        and s.get_member() == "VolumeUpdated":

        new_volume = 0
        for chan,vol in args[0]:
            if vol > new_volume:
                new_volume = vol

        if new_volume != userdata.watch_volume:
            alsa_value, alsa_db = get_amixer_value()
            userdata.watch_volume = new_volume

            print_volume(new_volume, alsa_value, alsa_db)

def sweep(start=0, stop=PA_NATIVE_0DB, step=userdata.sweep_step, full=False):

    vol(start)
    alsa_value, new_alsa_db = get_amixer_value()

    for vol_native in range(start, stop, step):
        set_volume_dbus(userdata.stream_restore_entry, vol_native)

        new_alsa_value, new_alsa_db = get_amixer_value()
        assert(new_alsa_value is not None)
        if full or not full and new_alsa_value != alsa_value:
            vol(vol_native)
            alsa_value = new_alsa_value

        if userdata.sweep_sleep is not None:
            time.sleep(userdata.sweep_sleep)

def set_sleep(sleep):
    userdata.sweep_sleep = sleep

def vol(value):
    stp(value, True)

def stp(step, absolute_step=False):
    new_vol = 0

    if type(step) != type(int()) and type(step) != type(float()):
        print "Argument is not valid. (integer value for changing sw volume, float for dB volume)"
        return

    if absolute_step:
        if type(step) == type(float()):
            new_vol = db2sw(step)
        else:
            new_vol = step
    else:
        current_vol = get_volume()
        if type(step) == type(int()):
            new_vol = new_vol + step
        # only update dB step if previous volume is above mute
        elif current_vol > 0:
            current_vol_dB = sw2db(current_vol)
            new_vol_dB = current_vol_dB + step
            new_vol = db2sw(new_vol_dB)

    if new_vol < 0:
        new_vol= 0
    s(new_vol)

if __name__ == "__main__":
    main()

